@@ -38,6 +38,7 @@ import utils.file_write as write
 import utils.file_search as search
 import utils.file_convert as convert
 import utils.file_image as image
+import utils.exiftool as exiftool
 # from utils.dir import get_folders_ftp
 # import utils.string_utils as strUtils
 
@@ -719,6 +720,7 @@ def get_data(file_path, **kwargs):
             if 'size' in data_include:
                 file_data['size'] = os.path.getsize(file_path)
 
+
             return file_data
         except FileNotFoundError as error:
             logger.warning("Error: %s", error)
@@ -1175,6 +1177,38 @@ def get_files(search_path=False, **kwargs):
 
         # path_files = index_files(path, extension_array, ignore_array, recursive)
         # file_array = path_files + file_array
+    file_array = _add_jpg_meta_data(file_array)
+    return file_array
+
+def _add_jpg_meta_data(file_array):
+    # print(f"_add_jpg_meta_data.file_array: {len(file_array)}")
+    result_array = []
+    jpg_array = []
+    jpg_path_array = []
+    for file in file_array:
+        if file['extension'] == ".jpg":
+            # print(f"found jpg file: {file['file_path']}")
+            file['file_path'] = strUtils.format.file_path(file['file_path'],url=True)
+            file['file_path_exif_copy'] = strUtils.format.file_path(f"{file['file_path']}_original",url=True)
+            jpg_array.append(file)
+            jpg_path_array.append(file['file_path'])
+        else:
+            result_array.append(file)
+    
+    if len(jpg_array) > 0:
+        # print(f"jpg_array: {len(jpg_array)}")
+        with exiftool.ExifTool(executable_=r"Z:\Structure\Archive\Programming Packages - Libraries\Python\colemen_file_utils\.venv\exiftool.exe") as et:
+            datas = et.get_metadata_batch(jpg_path_array)
+            for meta_data in datas:
+                if 'XMP:Subject' not in meta_data:
+                    meta_data['XMP:Subject'] = []
+                if 'IPTC:Keywords' not in meta_data:
+                    meta_data['IPTC:Keywords'] = []
+                for jpg in jpg_array:
+                    if jpg['file_path'] == meta_data['SourceFile']:
+                        jpg['meta_data'] = meta_data
+                        result_array.append(jpg)
+        return result_array
     return file_array
 
 def _gen_extension_array(ext_array):
@@ -1458,6 +1492,7 @@ class GetFilesThreaded:
         # print(f"                                                                                                                      ", end="\r", flush=True)
         for thread in self.thread_array:
             thread.join()
+        self.file_array = _add_jpg_meta_data(self.file_array)
         return self.file_array
 
         # def _get_files_single_thread(search_path=False, recursive=False, exclude=None, include=None, extensions=None, data_include=None):
