@@ -9,9 +9,10 @@
 # import json
 # import shutil
 # import os
-
-
 # import ffmpeg
+from distutils import extension
+import time
+import os
 
 
 import json
@@ -31,6 +32,30 @@ REGEX_TYPE = type(re.compile('hello, world'))
 # from threading import Thread
 
 def _parse_file_obj_list(file):
+    '''
+        The "file" provided is parsed to into an array of file objects.
+        "file" being a super general term for a file_path, list of paths, list of dicts with "file_path" prop.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `file` {string|list|dict}
+            A file path, list of paths, a file_object dictionary or list of dictionaries.
+
+        Return {list}
+        ----------------------
+        A list of file_objects (created by file.get_data())
+        If nothing is found the list is empty.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 01-27-2022 10:24:49
+        `memberOf`: file_image
+        `version`: 1.0
+        `method_name`: _parse_file_obj_list
+    '''
     file_array = []
     if isinstance(file, (str)):
         if f.exists(file):
@@ -110,9 +135,9 @@ def scale(src_path,size,**kwargs):
         else:
             sizeTuple = (height, height)
             
-    fileData = f.get_data(src_path)
+    file_data = f.get_data(src_path)
     if dst_path is False:
-        dst_path = f"{fileData['dir_path']}/{fileData['name_no_ext']}_{sizeTuple[0]}x{sizeTuple[1]}{fileData['extension']}"
+        dst_path = f"{file_data['dir_path']}/{file_data['name_no_ext']}_{sizeTuple[0]}x{sizeTuple[1]}{file_data['extension']}"
     
     image = Image.open(src_path)
     image.thumbnail(sizeTuple, Image.ANTIALIAS)
@@ -128,6 +153,7 @@ def delete_all_keywords(files):
 
 def delete_keyword(files,keywords='*',**kwargs):
     case_sensitive = obj.get_kwarg(['case sensitive'],True,(bool),**kwargs)
+    save = obj.get_kwarg(['save'],False,(bool),**kwargs)
     needle_array = _keywords_to_list(keywords)
     update_array = []
     if len(needle_array) == 0:
@@ -142,7 +168,7 @@ def delete_keyword(files,keywords='*',**kwargs):
         kws = get_keywords(file)
         
         new_keys = []
-        print(f"kws: ",kws)
+        # print(f"kws: ",kws)
         for haystack in kws:
             haystack = f"{haystack}"
             for needle in needle_array:
@@ -163,17 +189,19 @@ def delete_keyword(files,keywords='*',**kwargs):
         if len(new_keys) > 1:
             new_keys = list(set(new_keys))
         if len(kws) != len(new_keys):
-            print(f"total original keys: {len(kws)}")
-            print(f"total new_keys: {len(new_keys)}")
+            # print(f"total original keys: {len(kws)}")
+            # print(f"total new_keys: {len(new_keys)}")
             file['meta_data']['XMP:Subject'] = new_keys
             file['meta_data']['IPTC:Keywords'] = new_keys
             update_array.append(file)
-    save_file_obj(update_array)
-    return update_array
+    if save is True:
+        save_file_obj(update_array)
+    return file_array
             # purge_original(file)
         
 def add_keyword(files,keywords='',**kwargs):
     snake_case = obj.get_kwarg(['snake case'],True,(bool),**kwargs)
+    save = obj.get_kwarg(['save'],False,(bool),**kwargs)
     
     file_array = _parse_file_obj_list(files)
     # print(f"add_keyword.keywords: ",keywords)
@@ -191,10 +219,45 @@ def add_keyword(files,keywords='',**kwargs):
         if len(kws) != len(new_keys):
             update_array.append(file)
 
-    save_file_obj(update_array)  
-    return update_array
+    if save is True:
+        save_file_obj(update_array)
+    return file_array
 
 def set_keywords(files,keywords='',**kwargs):
+    '''
+        Sets the keyword values on the file(s) provided.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `files` {list|string|dict}
+            A file path, list of paths, file object dictionary or list of dictionaries.
+            
+        `keywords` {list|string}
+            A string of comma separated keywords or a list of keywords to add to the file.
+
+        Keyword Arguments
+        -------------------------
+        [`save`=False] {bool}
+            if True, the file is saved after its keywords are updated.
+            
+        [`snake_case`=False] {bool}
+            format the keywords to be in snake case notation.
+
+        Return {list}
+        ----------------------
+        The list of files with updated keywords.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 01-27-2022 11:08:05
+        `memberOf`: file_image
+        `version`: 1.0
+        `method_name`: set_keywords
+    '''
+    save = obj.get_kwarg(['save'],False,(bool),**kwargs)
     snake_case = obj.get_kwarg(['snake case'],True,(bool),**kwargs)
     
     file_array = _parse_file_obj_list(files)
@@ -207,15 +270,10 @@ def set_keywords(files,keywords='',**kwargs):
         file['meta_data']['IPTC:Keywords'] = keywords
         # exif_tool_array.append([file['file_path'],keywords])
 
-    save_file_obj(files)
-    # with exiftool.ExifTool(executable_=r"Z:\Structure\Archive\Programming Packages - Libraries\Python\colemen_file_utils\.venv\exiftool.exe") as et:
-        # for file in file_array:
-            # print(f"set_keywords.filePath: {file[0]}")
-            # et.set_tags({"XMP:Subject":file[1]},[file[0]])
-            # et.set_tags({"IPTC:Keywords":file[1]},[file[0]])      
-            # purge_original(file[0])
-            
-        # purge_original(file)
+    if save is True:
+        save_file_obj(files)
+
+    return file_array
 
 def save_file_obj(files):
     file_array = _parse_file_obj_list(files)
@@ -225,10 +283,6 @@ def save_file_obj(files):
             if 'meta_data' in file:
                 print(f"save_file_obj.filePath: {file['file_path']}")
                 et.set_tags(file['meta_data'],file['file_path'])
-                # f.write.to_json("imgs.json",file)
-                # exit()
-                # for k,v in file['meta_data'].items():
-                    # et.set_tags(file['meta_data'],file['file_path'])
 
 def tags_to_snakecase(files):
     file_array = _parse_file_obj_list(files)
@@ -358,18 +412,52 @@ def get_keywords(files):
     return keywords
 
 def replace_keyword(files,needle,replace):
+    '''
+        Find and replace keyword(s) in the files.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `files` {list|string|dict}
+            A file path, list of paths, file object dictionary or list of dictionaries.
+            
+        `needle` {list|string}
+            The keyword or list of keywords to replace. 
+
+        `replace` {string}
+            What to replace the keyword with.
+
+        Return {list}
+        ----------------------
+        The list of files with updated keywords.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 01-27-2022 11:12:23
+        `memberOf`: file_image
+        `version`: 1.0
+        `method_name`: replace_keyword
+    '''
     if isinstance(needle,(list)) is False:
         needle = [needle]
     file_array = _parse_file_obj_list(files)
+    update_array = []
     for file in file_array:
         if 'meta_data' not in file:
             file = get_meta(file)
         kws = get_keywords(file)
-        for x in needle:
-            if x in kws:
-                kws.remove(x)
+        for ndl in needle:
+            if ndl in kws:
+                kws.remove(ndl)
                 kws.append(replace)
                 set_keywords(file,kws)
+                update_array.append(file)
+    if len(update_array) > 0:
+        save_file_obj(update_array)
+        
+    return file_array
 
 def has_keyword(files,keywords,**kwargs):
     # Causes the search to reverse, so if the image does NOT have a tag, it is returned
@@ -410,7 +498,172 @@ def has_keyword(files,keywords,**kwargs):
     #     return False
     # if we are searching multiple files, we return an array of files with the keywords
     return result_array
+
+def apply_keyword_synonyms(files,synonyms,**kwargs):
+    '''
+        Iterates the files provided, and applies synonyms if a matching key is found.
+
+        ----------
+
+        Arguments
+        -------------------------
+        `files` {list|string|dict}
+            A file path, list of paths, file object dictionary or list of dictionaries.
+        `synonyms` {dict}
+            The synonym object should be organized like this:
+            
+            {'key':['synonym1','synonym2']}
+            
+
+        Return {list}
+        ----------------------
+        The list of files provided, with the synonyms added.
+
+        Meta
+        ----------
+        `author`: Colemen Atwood
+        `created`: 01-27-2022 10:43:23
+        `memberOf`: file_image
+        `version`: 1.0
+        `method_name`: apply_keyword_synonyms
+    '''
+    save = obj.get_kwarg(['save'],True,(bool),**kwargs)
+    # Prepare the files array for iteration and gather necessary info.
+    files_array = _parse_file_obj_list(files)
+    update_array = []
+    key_synonyms = synonyms
+    update_array = []
+    
+    # If there are no files, just return the empty array.
+    if len(files_array) == 0:
+        return files_array
+    
+    for file in files_array:
+        kws = f.image.get_keywords(file)
+        kws_hash_prime = csu.gen.hash(kws)
+        if isinstance(kws,(list)) is False:
+            kws = [kws]
+        new_tags = kws
+        for key in kws:
+            # print(f"key: ",key)
+            if isinstance(key,(str)) is True:
+                if key in key_synonyms:
+                    for k in key_synonyms[key]:
+                        if k not in kws:
+                            new_tags.append(k)
+                        
+        file['meta_data']['XMP:Subject'] = new_tags
+        file['meta_data']['IPTC:Keywords'] = new_tags
         
+        print(f"kws_hash_prime: {kws_hash_prime}")
+        # kws = cfu.file.image.get_keywords(file)
+        kwsHash = csu.gen.hash(new_tags)
+        print(f"kwsHash:        {kwsHash}")
+        
+        # compare the original keyword hash to the new one
+        # if they are not identical, that means new keywords
+        # were added, so we add it to the update_array.
+        if kws_hash_prime != kwsHash:
+            print(f"new tags found.")
+            # file['meta_data']['XMP:Subject'] = kws
+            # file['meta_data']['IPTC:Keywords'] = kws
+            update_array.append(file)
+            # print(f"file['meta_data']['XMP:Subject']: ",file['meta_data']['XMP:Subject'])
+            # exit()
+            
+    # if any files need updating, then we save the data to the file..
+    if len(update_array) > 0 and save is True:
+        f.image.save_file_obj(update_array)
+    return files_array
+
+def auto_tag_by_folder_name(watch_path,**kwargs):
+    print(f"auto_tag_by_folder_name")
+    synonyms = obj.get_kwarg(['synonyms'],False,(bool,dict),**kwargs)
+    all_files = f.get_files(watch_path, ext="images")
+    print(f"watching {len(all_files)} files")
+    # duration = 0
+
+    while True:
+        time.sleep(.5)
+        if len(all_files) == 0:
+            all_files = f.get_files(watch_path, ext="images")
+        # duration += .5
+        # if duration % 10 == 0:
+        #     transferCompleteImages.transferCompleteImages()
+        #     transferDLToOrg()
+        #     convertOrgFolder()
+        moved_files = []
+        new_all_files = []
+        update_files = []
+        
+        
+        for file in all_files:
+            # if the file does not exist, append it to the moved_files list.
+            if f.exists(file['file_path']) is False:
+                print(f"moved: {file['file_name']}")
+                moved_files.append(file)
+            else:
+                new_all_files.append(file)
+
+        for file in moved_files:
+            # search for the file by its name
+            result = f.search.by_name(file['file_name'], watch_path)
+            if result is not None:
+                print(f"result: {result}")
+                # get the file data, essentially update its paths and such.
+                file_data = f.get_data(result[0])
+                # append the file to the update_files array
+                update_files.append(file_data)
+        
+        new_files = _tag_by_folder_name(update_files,os.path.basename(watch_path),save=False)
+        if synonyms is not False:
+            new_files = apply_keyword_synonyms(new_files,synonyms,save=False)
+        save_file_obj(new_files)
+        
+        new_all_files.extend(new_files)
+        all_files = new_all_files    
+    
+def _tag_by_folder_name(files,exclude='',**kwargs):
+    save = obj.get_kwarg(['save'],False,(bool),**kwargs)
+    if isinstance(exclude,(str)):
+        exclude = [exclude]
+    files_array = _parse_file_obj_list(files)
+    update_array = []
+    # new_file_array = []
+    for file in files_array:
+        # parse the folder name from the file object
+        dir_name = os.path.basename(file['dir_path'])
+        if dir_name not in exclude:
+            file = add_keyword(file,dir_name,save=False)[0]
+            update_array.append(file)
+            
+            
+        # new_file_array.append(file)
+    # new_file_array = applyKeywordSynonyms(new_file_array)
+    # update_files(new_file_array)
+    if save is True:
+        save_file_obj(update_array)
+    return files_array
+
+def tag_if_missing_keyword(files,needles,keyword):
+    update_array = []
+    files_array = _parse_file_obj_list(files)
+    print(f"")
+    # get all of the files with a needle in their keywords.
+    has_needle = has_keyword(files_array,needles)
+    # get all of the files that already have the "keyword" and a needle
+    missingNeedle = has_keyword(has_needle,keyword)
+    if len(missingNeedle) > 0:
+        # remove the "keyword" from those files.
+        update_array.extend(delete_keyword(missingNeedle,keyword))
+    
+    # get all of the files that do NOT have a needle in their keywords
+    result = has_keyword(files_array,needles,reverse=True)
+    # add the "keyword" to them
+    update_array.extend(add_keyword(result,keyword))
+    save_file_obj(update_array)
+
+
 # def has_tag(file,tag):
 #     result_array = []
 #     tags_array = []
